"""Tests for test mocking utilities."""

import time

from theia.core import test_mocks


def test_mock_time_simple() -> None:
    ts = time.perf_counter()

    with test_mocks.mock_time(0):
        assert time.perf_counter() == 0
        assert time.perf_counter() != ts

    assert ts < time.perf_counter() < ts + 1


def test_mock_time_nested() -> None:
    with test_mocks.mock_time(0):
        assert time.perf_counter() == 0

        with test_mocks.mock_time(1):
            assert time.perf_counter() == 1

        assert time.perf_counter() == 0



import collections
import time
from typing import Deque, Generic, Iterable, TypeVar

T = TypeVar("T")
U = TypeVar("U")


class Timer:
    """A timer that can be used to measure the duration of an operation."""

    def __init__(self, duration_s: float | None = None):
        self._start_time = _now()
        self._duration_s = duration_s

    def elapsed_s(self) -> float:
        return _now() - self._start_time

    def is_finished(self) -> bool:
        if self._duration_s is None:
            raise ValueError("Timer has no duration.")
        return self.elapsed_s() >= self._duration_s


class ExpiringSet(Generic[T]):
    """A set whose entries expire after a configured duration."""

    def __init__(self, expiration_s: float):
        super().__init__()
        self._expiration_s = expiration_s
        self._timed_keys: dict[T, float] = {}

    def _discard_expired(self) -> None:
        """Discards expired entries from the dictionary."""
        # Since dict iteration follows insertion order, the oldest values will be
        # encountered first.
        now = _now()
        expired_keys = []
        for key, timestamp in self._timed_keys.items():
            if now - timestamp >= self._expiration_s:
                expired_keys.append(key)
            else:
                break

        for key in expired_keys:
            del self._timed_keys[key]

    def add(self, key: T) -> None:
        self._discard_expired()
        # Remove the object before adding such that insertion order corresponds with age.
        if key in self._timed_keys:
            del self._timed_keys[key]
        self._timed_keys[key] = _now()

    def values(self) -> set[T]:
        self._discard_expired()
        return set(self._timed_keys)

    def contains(self, key: T) -> bool:
        self._discard_expired()
        return key in self._timed_keys


class ExpiringDict(Generic[T, U]):
    """A dictionary whose entries expire after a configured duration."""

    def __init__(self, expiration_s: float):
        super().__init__()
        self._expiration_s = expiration_s
        self._timed_items: dict[T, tuple[float, U]] = {}

    def _discard_expired(self) -> None:
        """Discards expired entries from the dictionary."""
        # Since dict iteration follows insertion order, the oldest values will be
        # encountered first.
        now = _now()
        expired_keys = []
        for key, (timestamp, _) in self._timed_items.items():
            if now - timestamp >= self._expiration_s:
                expired_keys.append(key)
            else:
                break

        for key in expired_keys:
            del self._timed_items[key]

    def set(self, key: T, value: U) -> None:
        # Remove the object before adding such that insertion order corresponds with age.
        if key in self._timed_items:
            del self._timed_items[key]
        self._discard_expired()
        self._timed_items[key] = (_now(), value)

    def items(self) -> Iterable[tuple[T, U]]:
        self._discard_expired()
        return ((k, v) for k, (_, v) in self._timed_items.items())

    def get(self, key: T) -> U | None:
        self._discard_expired()
        item = self._timed_items.get(key)
        if item is None:
            return None
        _, value = item
        return value

    def delete(self, key: T) -> None:
        if key in self._timed_items:
            del self._timed_items[key]


class ExpiringQueue(Generic[T]):
    """A queue whose values expire after a configured duration."""

    def __init__(self, expiration_s: float):
        self._expiration_s = expiration_s
        self._timed_values: Deque[tuple[float, T]] = collections.deque()

    def _discard_expired(self) -> None:
        """Discards expired values from the queue."""
        now = _now()
        while self._timed_values:
            then, _ = self._timed_values[0]
            if now - then >= self._expiration_s:
                self._timed_values.popleft()
            else:
                break

    def append(self, obj: T) -> None:
        # Even though this method does not access the queue, we need to discard expired
        # values to ensure the queue does not grow without bound.
        self._discard_expired()
        self._timed_values.append((_now(), obj))

    def values(self) -> list[T]:
        self._discard_expired()
        return [obj for _, obj in self._timed_values]


class FrequencyCounter:
    """Tracks event frequency within a specified time window."""

    def __init__(self, window_s: float):
        self._window_s = window_s
        self._expiring_q: ExpiringQueue[None] = ExpiringQueue(window_s)

    def update(self) -> None:
        """Records a new event at the current time."""
        self._expiring_q.append(None)

    def frequency(self) -> float:
        """Returns the current event frequency."""
        q_values = self._expiring_q.values()
        return len(q_values) / self._window_s


def _now() -> float:
    return time.perf_counter()



import asyncio
import ipaddress
from typing import AsyncGenerator

import aiohttp

from theia.core import primitives, timing, web

_DEFAULT_TIMEOUT_S = 2
_DEFAULT_TIMEOUT = aiohttp.ClientTimeout(total=_DEFAULT_TIMEOUT_S)


class Client:
    """A basic client for interacting with the Theia server.

    Offers basic asyncio-based HTTP and WebSocket connections to the server.
    """

    def __init__(self, host: ipaddress.IPv4Address | str, port: int):
        self._base_url = f"http://{host}:{port}"
        self._session: aiohttp.ClientSession | None = None

    async def open(self, timeout: float | None = None) -> None:
        """Opens an HTTP session linking the client to the server.

        By default, this will wait until the server is ready. If a timeout is provided,
        it will raise an aiohttp.ClientError if the server is not reachable within the
        given timeout.
        """
        if self._session:
            return
        await self._wait_for_server_ready(timeout)
        self._session = aiohttp.ClientSession(self._base_url, raise_for_status=True)

    async def close(self) -> None:
        """Closes the HTTP session linking the client to the server."""
        if self._session:
            await self._session.close()
            self._session = None

    async def status_stream(self) -> AsyncGenerator[primitives.SystemStatus, None]:
        """Establishes a websocket connection and yields realtime status updates.

        Raises a aiohttp.client_exceptions.WSMessageTypeError if the server shuts down
        while the client is still connected.
        """
        if not self._session:
            raise RuntimeError("Client is not open.")

        path = web.status_path()
        async with self._session.ws_connect(path) as ws:
            while not ws.closed:
                raw_status = await ws.receive_json()
                yield primitives.SystemStatus.from_dict(raw_status)

    async def get_image(
        self, stream_id: str, pts: int, draw: bool | None = None
    ) -> bytes:
        """Gets the image corresponding to the given pts.

        Specify draw to True in order to include the frame's bounding boxes on the
        image. Otherwise the image will be returned as is.

        Raises a RuntimeError if the client is not open.
        """
        if not self._session:
            raise RuntimeError("Client is not open.")

        path = web.stream_image_path(stream_id, pts)
        params: dict[str, str] = {}
        if draw is not None:
            params["draw"] = str(draw).lower()
        async with self._session.get(path, params=params) as resp:
            resp.raise_for_status()
            image = await resp.read()
            return image

    async def _get_ready(self) -> None:
        """Check if the server is ready. Does NOT require that the client is open.

        Raises an aiohttp.ClientError if the server is not ready to accept connections.
        """
        full_path = self._base_url + web.ready_path()
        # Uses the aiohttp basic API because the session may not be open yet.
        async with aiohttp.request("GET", full_path) as resp:
            resp.raise_for_status()
            if resp.status != 200:
                raise RuntimeError("Encountered unexpected response from server!")
            return

    async def _wait_for_server_ready(self, timeout: float | None) -> None:
        """Waits until the server is ready.

        Raises an aiohttp.ClientError if the server is not reachable within the given
        timeout.
        """
        timer = timing.Timer(timeout)
        while True:
            try:
                await self._get_ready()
                break
            except aiohttp.ClientError:
                if timeout is not None and timer.is_finished():
                    raise
                await asyncio.sleep(1)



import dataclasses
import pathlib

import yaml


@dataclasses.dataclass(kw_only=True)
class CameraConfig:
    resolution: tuple[int, int]
    rtsp_url: str


@dataclasses.dataclass(kw_only=True)
class ServerConfig:
    cameras: dict[str, CameraConfig]
    close_bbox_diagonal: float
    http_port: int
    recording_lifetime_s: float
    status_frequency_hz: float

    @property
    def camera_ids(self) -> list[str]:
        return list(self.cameras.keys())


def load(path: pathlib.Path) -> ServerConfig:
    """Loads the server configuration from a YAML file."""
    with open(path, "r") as f:
        loaded_cfg = yaml.safe_load(f)
    cams = {
        cam_id: CameraConfig(
            # YAML doesn't have built in tuple support so we convert it here.
            resolution=tuple(cam_dict["resolution"]),
            rtsp_url=cam_dict["rtsp_url"],
        )
        for cam_id, cam_dict in loaded_cfg.get("cameras", {}).items()
    }
    server_cfg = ServerConfig(
        cameras=cams,
        close_bbox_diagonal=loaded_cfg["close_bbox_diagonal"],
        http_port=loaded_cfg["http_port"],
        recording_lifetime_s=loaded_cfg["recording_lifetime_s"],
        status_frequency_hz=loaded_cfg["status_frequency_hz"],
    )
    if len(server_cfg.cameras) == 0:
        raise ValueError("Configuration must specify at least one camera.")
    return server_cfg


def save(config: ServerConfig, path: pathlib.Path) -> None:
    """Saves the server configuration to a YAML file."""
    data = dataclasses.asdict(config)

    # YAML doesn't have built in tuple support so we convert it here.
    for cam_cfg in data["cameras"].values():
        cam_cfg["resolution"] = list(cam_cfg["resolution"])

    with open(path, "w") as f:
        yaml.safe_dump(data, f)



import math

import pytest

from theia.core import test_mocks, timing

_TIMEOUT_S = 10


def test_timer_no_duration() -> None:
    with test_mocks.mock_time(0):
        timer = timing.Timer()
    with test_mocks.mock_time(1):
        assert timer.elapsed_s() == 1.0
    with pytest.raises(ValueError):
        timer.is_finished()


def test_timer_duration() -> None:
    with test_mocks.mock_time(0):
        timer = timing.Timer(_TIMEOUT_S)
        assert not timer.is_finished()
    with test_mocks.mock_time(_TIMEOUT_S / 2):
        assert not timer.is_finished()
    with test_mocks.mock_time(2 * _TIMEOUT_S):
        assert timer.is_finished()


def test_frequency_counter_not_started() -> None:
    """Test that FrequencyCounter is not ready and returns 0 for rate and count if
    update is never called.
    """
    counter = timing.FrequencyCounter(_TIMEOUT_S)
    with test_mocks.mock_time(0):
        assert counter.frequency() == 0
    with test_mocks.mock_time(_TIMEOUT_S * 2):
        assert counter.frequency() == 0


def test_frequency_counter_update() -> None:
    counter = timing.FrequencyCounter(_TIMEOUT_S)
    with test_mocks.mock_time(0):
        counter.update()
    with test_mocks.mock_time(0.25 * _TIMEOUT_S):
        freq = counter.frequency()
        expected_freq = 1 / _TIMEOUT_S
        assert math.isclose(freq, expected_freq, abs_tol=1e-5)
        counter.update()
    with test_mocks.mock_time(0.75 * _TIMEOUT_S):
        freq = counter.frequency()
        expected_freq = 2 / _TIMEOUT_S
        assert math.isclose(freq, expected_freq, abs_tol=1e-5)
        counter.update()
    with test_mocks.mock_time(_TIMEOUT_S):
        freq = counter.frequency()
        expected_freq = 2 / _TIMEOUT_S
        assert math.isclose(freq, expected_freq, abs_tol=1e-5)


def test_expiring_queue_expiration() -> None:
    with test_mocks.mock_time(0):
        expiring_q = timing.ExpiringQueue[int](2)
    with test_mocks.mock_time(1):
        expiring_q.append(1)
        expiring_q.append(2)
        assert expiring_q.values() == [1, 2]
    with test_mocks.mock_time(2):
        expiring_q.append(1)
        assert expiring_q.values() == [1, 2, 1]
    with test_mocks.mock_time(3):
        assert expiring_q.values() == [1]
    with test_mocks.mock_time(4):
        assert expiring_q.values() == []


def test_expiring_dict_expiration() -> None:
    with test_mocks.mock_time(0):
        expiring_dict = timing.ExpiringDict[str, int](2)
    with test_mocks.mock_time(1):
        expiring_dict.set("A", 1)
        expiring_dict.set("B", 2)
        assert dict(expiring_dict.items()) == {"A": 1, "B": 2}
    with test_mocks.mock_time(2):
        expiring_dict.set("A", 3)
        assert dict(expiring_dict.items()) == {"A": 3, "B": 2}
    with test_mocks.mock_time(3):
        assert dict(expiring_dict.items()) == {"A": 3}
    with test_mocks.mock_time(4):
        assert dict(expiring_dict.items()) == {}
    with test_mocks.mock_time(5):
        expiring_dict.set("A", 1)
        expiring_dict.delete("B")
        assert dict(expiring_dict.items()) == {"A": 1}
        expiring_dict.delete("A")
        assert dict(expiring_dict.items()) == {}


def test_expiring_set_expiration() -> None:
    with test_mocks.mock_time(0):
        expiring_set = timing.ExpiringSet[int](2)
    with test_mocks.mock_time(1):
        expiring_set.add(1)
        expiring_set.add(2)
        assert expiring_set.values() == {1, 2}
    with test_mocks.mock_time(2):
        expiring_set.add(1)
        assert expiring_set.values() == {1, 2}
    with test_mocks.mock_time(3.5):
        assert expiring_set.values() == {1}
